##############################################################################
#
# Copyright (c) 2017, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest
from io import StringIO

from ZConfig import validator
from ZConfig.tests import support


def run_validator(*args):
    return validator.main(args)


class TestValidator(unittest.TestCase):

    def test_no_schema(self):
        sio = StringIO()
        with support.stderr_replaced(sio):
            with self.assertRaises(SystemExit) as cm:
                run_validator()
        self.assertEqual(cm.exception.code, 2)
        err = sio.getvalue()
        # Checked separately since these are included very differently
        # with different versions of Python's argparse module.
        self.assertIn('-s/--schema', err)
        self.assertIn(' required', err)

    def test_schema_only(self):
        res = run_validator("--schema", support.input_file('simple.xml'))
        self.assertEqual(res, 0)

    @support.with_stdin_from_input_file('simple.conf')
    def test_schema_only_redirect(self):
        res = run_validator("--schema", support.input_file('simple.xml'))
        self.assertEqual(res, 0)

    def test_good_config(self):
        res = run_validator("--schema", support.input_file('simple.xml'),
                            support.input_file('simple.conf'),
                            support.input_file('simple.conf'))
        self.assertEqual(res, 0)

    def test_bad_config(self):
        sio = StringIO()
        with support.stderr_replaced(sio):
            res = run_validator("--schema", support.input_file("simple.xml"),
                                support.input_file("outer.conf"))
        self.assertEqual(res, 1)
        self.assertIn("'refouter' is not a known key name", sio.getvalue())


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2017, 2018 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from contextlib import contextmanager


try:
    import docutils.frontend
    import docutils.parsers.rst
    import docutils.utils
    from docutils import nodes
    from docutils.parsers.rst import Directive
except ModuleNotFoundError:  # pragma: no cover
    RstSchemaPrinter = None
    RstSchemaFormatter = None
else:

    from ZConfig._schema_utils import MARKER
    from ZConfig._schema_utils import AbstractSchemaFormatter
    from ZConfig._schema_utils import AbstractSchemaPrinter
    from ZConfig._schema_utils import load_schema

    class RstSchemaFormatter(AbstractSchemaFormatter):

        settings = None

        def __init__(self, schema, stream=None):
            super().__init__(schema, stream)
            self.document = None
            self._current_node = None
            self._nodes = []
            self.settings = docutils.frontend.OptionParser(
                components=(docutils.parsers.rst.Parser,)).get_default_values()

        def esc(self, text):
            return text

        def _parsed(self, text, name='Schema'):
            document = docutils.utils.new_document(
                name,
                settings=self.settings)

            parser = docutils.parsers.rst.Parser()
            parser.parse(text, document)
            return document.children

        def write(self, *texts):
            for text in texts:
                if isinstance(text, str):
                    self._current_node += nodes.Text(' ' + text + ' ', text)
                else:
                    # Already parsed
                    self._current_node += text

        def description(self, text):
            if not text:
                return

            self.write(self._parsed(self._dedent(text), "description"))

        def example(self, text):
            if not text:
                return

            dedented = self._dedent(text)
            example = "Example::\n\n\t" + '\n\t'.join(dedented.split('\n'))
            self.write(self._parsed(example, "example"))

        @contextmanager
        def item_list(self):
            old_node = self._current_node
            self._current_node = nodes.definition_list()
            old_node += self._current_node
            yield
            self._current_node = old_node

        @contextmanager
        def describing(self, description=MARKER, after=None):
            dl = self._current_node
            assert isinstance(dl, nodes.definition_list), dl
            item = nodes.definition_list_item()
            dl += item
            term = nodes.term()
            item += term
            self._current_node = term

            yield

            # We must now have either a description (so we call
            # described_as) or they must call described_as
            # des
            self._current_node = item

            self._describing(description, after)

        @contextmanager
        def described_as(self):
            item = self._current_node
            assert isinstance(item, nodes.definition_list_item), item

            definition = nodes.definition()
            para = nodes.paragraph()
            definition += para
            item += definition
            self._current_node = para

            yield

            # When this is done, we're back to the list
            self._current_node = item.parent

        def abstract_name(self, name):
            self._current_node += nodes.emphasis(text=name, rawsource=name)

        def concrete_name(self, *name):
            name = ' '.join(name)
            self._current_node += nodes.strong(text=name, rawsource=name)

        def concrete_section_name(self, *name):
            name = ' '.join(name)
            return self.concrete_name("<" + name + ">")

        @contextmanager
        def body(self):
            self.document = self._current_node = docutils.utils.new_document(
                "Schema",
                settings=self.settings)
            yield

    class RstSchemaPrinter(AbstractSchemaPrinter):
        _schema_formatter = RstSchemaFormatter

        def printSchema(self):
            super().printSchema()
            print(self.fmt.document.pformat(), file=self.fmt.stream)

    class SchemaToRstDirective(Directive):
        required_arguments = 1
        optional_arguments = 0
        option_spec = {
            'file': str,
            'members': str,
            'excluded-members': str,
        }

        def run(self):
            schema = load_schema(
                self.options.get('file', 'component.xml'),
                self.arguments[0],
            )

            members = ()
            if 'members' in self.options:
                members = self.options['members'].split()

            excluded_members = ()
            if 'excluded-members' in self.options:
                excluded_members = self.options['excluded-members'].split()

            printer = RstSchemaPrinter(schema, allowed_names=members,
                                       excluded_names=excluded_members)
            printer.fmt.settings = self.state.document.settings

            printer.buildSchema()

            return printer.fmt.document.children

    def setup(app):  # pragma: no cover
        "Sphinx extension entry point to add the zconfig directive."
        app.add_directive("zconfig", SchemaToRstDirective)



##############################################################################
#
# Copyright (c) 2002, 2003, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Tests of standard ZConfig datatypes."""

import datetime
import os
import shutil
import socket
import sys
import tempfile
import unittest

import ZConfig.datatypes
from ZConfig.tests.support import TestHelper


here = os.path.abspath(__file__)


class DatatypeTestCase(unittest.TestCase):
    types = ZConfig.datatypes.Registry()

    def test_datatype_basickey(self):
        convert = self.types.get("basic-key")
        eq = self.assertEqual
        raises = self.assertRaises

        eq(convert("abc"), "abc")
        eq(convert("ABC_DEF.123"), "abc_def.123")
        eq(convert("Abc-Def-456"), "abc-def-456")
        eq(convert("Abc.Def"), "abc.def")

        raises(ValueError, convert, "_abc")
        raises(ValueError, convert, "-abc")
        raises(ValueError, convert, "123")
        raises(ValueError, convert, "")

    def test_datatype_boolean(self):
        convert = self.types.get("boolean")
        check = self.assertTrue
        raises = self.assertRaises

        check(convert("on"))
        check(convert("true"))
        check(convert("yes"))
        check(not convert("off"))
        check(not convert("false"))
        check(not convert("no"))
        raises(ValueError, convert, '0')
        raises(ValueError, convert, '1')
        raises(ValueError, convert, '')
        raises(ValueError, convert, 'junk')

    def test_datatype_float(self):
        convert = self.types.get("float")
        eq = self.assertEqual
        raises = self.assertRaises

        eq(convert("1"), 1.0)
        self.assertIs(type(convert(1)), float)
        eq(convert("1.1"), 1.1)
        eq(convert("50.50"), 50.50)
        eq(convert("-50.50"), -50.50)
        eq(convert(0), 0.0)
        eq(convert("0"), 0.0)
        eq(convert("-0"), 0.0)
        eq(convert("0.0"), 0.0)

        raises(ValueError, convert, "junk")
        raises(ValueError, convert, "0x234.1.9")
        raises(ValueError, convert, "0.9-")

        # float handles inf/nan. Make sure conversion does too.
        for literal in ("inf", "-inf", b"inf", b"-inf"):
            eq(convert(literal), float(literal))

        # notably, nan is not equal to itself
        self.assertNotEqual(convert("nan"), float("nan"))
        self.assertNotEqual(convert(b"nan"), float(b"nan"))

    def test_datatype_identifier(self):
        convert = self.types.get("identifier")
        raises = self.assertRaises
        self.check_names(convert)
        self.check_never_namelike(convert)
        raises(ValueError, convert, ".abc")

    def check_names(self, convert):
        eq = self.assert_ascii_equal
        eq(convert, "AbcDef")
        eq(convert, "a________")
        eq(convert, "abc_def")
        eq(convert, "int123")
        eq(convert, "_abc")
        eq(convert, "_123")
        eq(convert, "__dict__")

    def assert_ascii_equal(self, convert, value):
        v = convert(value)
        self.assertEqual(v, value)
        self.assertIsInstance(v, str)

    def check_never_namelike(self, convert):
        raises = self.assertRaises
        raises(ValueError, convert, "2345")
        raises(ValueError, convert, "23.45")
        raises(ValueError, convert, ".45")
        raises(ValueError, convert, "23.")
        raises(ValueError, convert, "abc.")
        raises(ValueError, convert, "-abc")
        raises(ValueError, convert, "-123")
        raises(ValueError, convert, "abc-")
        raises(ValueError, convert, "123-")
        raises(ValueError, convert, "-")
        raises(ValueError, convert, ".")
        raises(ValueError, convert, "&%$*()")
        raises(ValueError, convert, "")

    def test_datatype_dotted_name(self):
        convert = self.types.get("dotted-name")
        raises = self.assertRaises
        self.check_names(convert)
        self.check_dotted_names(convert)
        self.check_never_namelike(convert)
        raises(ValueError, convert, "abc.")
        raises(ValueError, convert, ".abc.")
        raises(ValueError, convert, "abc.def.")
        raises(ValueError, convert, ".abc.def.")
        raises(ValueError, convert, ".abc.def")

    def test_datatype_dotted_suffix(self):
        convert = self.types.get("dotted-suffix")
        eq = self.assert_ascii_equal
        raises = self.assertRaises
        self.check_names(convert)
        self.check_dotted_names(convert)
        self.check_never_namelike(convert)
        eq(convert, ".a")
        eq(convert, ".a.b")
        eq(convert, ".a.b.c.d.e.f.g.h.i.j.k.l.m.n.o")
        raises(ValueError, convert, "abc.")
        raises(ValueError, convert, ".abc.")
        raises(ValueError, convert, "abc.def.")
        raises(ValueError, convert, ".abc.def.")

    def check_dotted_names(self, convert):
        eq = self.assert_ascii_equal
        eq(convert, "abc.def")
        eq(convert, "abc.def.ghi")
        eq(convert, "a.d.g.g.g.g.g.g.g")

    def test_datatype_inet_address(self):
        convert = self.types.get("inet-address")
        eq = self.assertEqual
        defhost = ZConfig.datatypes.DEFAULT_HOST
        eq(convert("Host.Example.Com:80"), ("host.example.com", 80))
        eq(convert("Host.Example.Com:0"), ("host.example.com", 0))
        eq(convert(":80"), (defhost, 80))
        eq(convert("80"), (defhost, 80))
        eq(convert("[::1]:80"), ("::1", 80))
        eq(convert("host.EXAMPLE.com"), ("host.example.com", None))
        eq(convert("2001::ABCD"), ("2001::abcd", None))
        self.assertRaises(ValueError, convert, "40 # foo")

    def test_datatype_inet_binding_address(self):
        convert = self.types.get("inet-binding-address")
        eq = self.assertEqual
        defhost = ""
        eq(convert("Host.Example.Com:80"), ("host.example.com", 80))
        eq(convert(":80"), (defhost, 80))
        eq(convert("80"), (defhost, 80))
        eq(convert("host.EXAMPLE.com"), ("host.example.com", None))
        self.assertRaises(ValueError, convert, "40 # foo")

    def test_datatype_inet_connection_address(self):
        convert = self.types.get("inet-connection-address")
        eq = self.assertEqual
        defhost = "127.0.0.1"
        eq(convert("Host.Example.Com:80"), ("host.example.com", 80))
        eq(convert(":80"), (defhost, 80))
        eq(convert("80"), (defhost, 80))
        eq(convert("host.EXAMPLE.com"), ("host.example.com", None))
        self.assertRaises(ValueError, convert, "40 # foo")

    def test_datatype_integer(self):
        convert = self.types.get("integer")
        eq = self.assertEqual
        raises = self.assertRaises

        eq(convert('-100'), -100)
        eq(convert('-1'), -1)
        eq(convert('-0'), 0)
        eq(convert('0'), 0)
        eq(convert('1'), 1)
        eq(convert('100'), 100)
        eq(convert('65535'), 65535)
        eq(convert('65536'), 65536)

        raises(ValueError, convert, 'abc')
        raises(ValueError, convert, '-0xabc')
        raises(ValueError, convert, '')
        raises(ValueError, convert, '123 456')
        raises(ValueError, convert, '123-')

    def test_datatype_locale(self):
        convert = self.types.get("locale")
        # Python supports "C" even when the _locale module is not available
        self.assertEqual(convert("C"), "C")
        self.assertRaises(ValueError, convert, "locale-does-not-exist")

    def test_datatype_port(self):
        convert = self.types.get("port-number")
        eq = self.assertEqual
        raises = self.assertRaises

        raises(ValueError, convert, '-1')
        eq(convert('0'), 0)
        eq(convert('1'), 1)
        eq(convert('80'), 80)
        eq(convert('1023'), 1023)
        eq(convert('1024'), 1024)
        eq(convert('60000'), 60000)
        eq(convert('65535'), 0xffff)
        raises(ValueError, convert, '65536')

    def test_datatype_socket_address(self):
        convert = self.types.get("socket-address")
        AF_INET = socket.AF_INET
        AF_INET6 = socket.AF_INET6
        defhost = ZConfig.datatypes.DEFAULT_HOST

        def check(value, family, address):
            a = convert(value)
            self.assertEqual(a.family, family)
            self.assertEqual(a.address, address)

        check("Host.Example.Com:80", AF_INET, ("host.example.com", 80))
        check(":80", AF_INET, (defhost, 80))
        check("80", AF_INET, (defhost, 80))
        check("host.EXAMPLE.com", AF_INET, ("host.example.com", None))
        check("::1", AF_INET6, ("::1", None))
        check("[::]:80", AF_INET6, ("::", 80))
        a1 = convert("/tmp/var/@345.4")
        a2 = convert("/tmp/var/@345.4:80")
        self.assertEqual(a1.address, "/tmp/var/@345.4")
        self.assertEqual(a2.address, "/tmp/var/@345.4:80")
        if hasattr(socket, "AF_UNIX"):
            self.assertEqual(a1.family, socket.AF_UNIX)
            self.assertEqual(a2.family, socket.AF_UNIX)
        else:  # pragma: no cover
            self.assertIsNone(a1.family)
            self.assertIsNone(a2.family)

        convert = self.types.get('socket-binding-address')
        check(":80", AF_INET, (defhost, 80))

        convert = self.types.get('socket-connection-address')
        check(":80", AF_INET, ("127.0.0.1", 80))

    def test_ipaddr_or_hostname(self):
        convert = self.types.get('ipaddr-or-hostname')
        eq = self.assertEqual
        raises = self.assertRaises
        eq(convert('hostname'), 'hostname')
        eq(convert('hostname.com'), 'hostname.com')
        eq(convert('www.hostname.com'), 'www.hostname.com')
        eq(convert('HOSTNAME'), 'hostname')
        eq(convert('HOSTNAME.COM'), 'hostname.com')
        eq(convert('WWW.HOSTNAME.COM'), 'www.hostname.com')
        eq(convert('127.0.0.1'), '127.0.0.1')
        eq(convert('::1'), '::1')
        eq(convert('2001:DB8:1234:4567:89AB:cdef:0:1'),
           '2001:db8:1234:4567:89ab:cdef:0:1')
        eq(convert('2001:DB8:1234:4567::10.11.12.13'),
           '2001:db8:1234:4567::10.11.12.13')
        raises(ValueError, convert, '1hostnamewithleadingnumeric')
        raises(ValueError, convert, '255.255')
        raises(ValueError, convert, '12345678')
        raises(ValueError, convert, '999.999.999.999')
        raises(ValueError, convert, 'a!badhostname')
        raises(ValueError, convert, '2001:DB8:0123:4567:89AB:cdef:0:1:2')
        raises(ValueError, convert, '2001:DB8:0123:4567::10.11.12.13.14')

    def test_existing_directory(self):
        convert = self.types.get('existing-directory')
        eq = self.assertEqual
        raises = self.assertRaises
        eq(convert('.'), '.')
        eq(convert(os.path.dirname(here)), os.path.dirname(here))
        raises(ValueError, convert, tempfile.mktemp())

    def test_existing_file(self):
        convert = self.types.get('existing-file')
        eq = self.assertEqual
        raises = self.assertRaises
        eq(convert('.'), '.')
        eq(convert(here), here)
        raises(ValueError, convert, tempfile.mktemp())

    def test_existing_path(self):
        convert = self.types.get('existing-path')
        eq = self.assertEqual
        raises = self.assertRaises
        eq(convert('.'), '.')
        eq(convert(here), here)
        eq(convert(os.path.dirname(here)), os.path.dirname(here))
        raises(ValueError, convert, tempfile.mktemp())

    def test_existing_dirpath(self):
        convert = self.types.get('existing-dirpath')
        eq = self.assertEqual
        raises = self.assertRaises
        eq(convert('.'), '.')
        eq(convert(here), here)
        raises(ValueError, convert, '/a/hopefully/nonexistent/path')
        raises(ValueError, convert, here + '/bogus')

    def test_byte_size(self):
        eq = self.assertEqual
        raises = self.assertRaises
        convert = self.types.get('byte-size')
        eq(convert('128'), 128)
        eq(convert('128KB'), 128 * 1024)
        eq(convert('128MB'), 128 * 1024 * 1024)
        eq(convert('128GB'), 128 * 1024 * 1024 * 1024)
        raises(ValueError, convert, '128TB')
        eq(convert('128'), 128)
        eq(convert('128kb'), 128 * 1024)
        eq(convert('128mb'), 128 * 1024 * 1024)
        eq(convert('128gb'), 128 * 1024 * 1024 * 1024)
        raises(ValueError, convert, '128tb')

    def test_time_interval(self):
        eq = self.assertEqual
        raises = self.assertRaises
        convert = self.types.get('time-interval')
        eq(convert('120'), 120)
        eq(convert('120S'), 120)
        eq(convert('120M'), 120 * 60)
        eq(convert('120H'), 120 * 60 * 60)
        eq(convert('120D'), 120 * 60 * 60 * 24)
        raises(ValueError, convert, '120W')
        eq(convert('120'), 120)
        eq(convert('120s'), 120)
        eq(convert('120m'), 120 * 60)
        eq(convert('120h'), 120 * 60 * 60)
        eq(convert('120d'), 120 * 60 * 60 * 24)
        raises(ValueError, convert, '120w')

    def test_timedelta(self):
        eq = self.assertEqual
        raises = self.assertRaises
        convert = self.types.get('timedelta')
        eq(convert('4w'), datetime.timedelta(weeks=4))
        eq(convert('2d'), datetime.timedelta(days=2))
        eq(convert('7h'), datetime.timedelta(hours=7))
        eq(convert('12m'), datetime.timedelta(minutes=12))
        eq(convert('14s'), datetime.timedelta(seconds=14))
        eq(convert('4w 2d 7h 12m 14s'),
           datetime.timedelta(2, 14, minutes=12, hours=7, weeks=4))

        raises(TypeError, convert, '1y')


class RegistryTestCase(TestHelper, unittest.TestCase):

    def test_registry_does_not_mask_toplevel_imports(self):
        old_sys_path = sys.path[:]
        tmpdir = tempfile.mkdtemp(prefix="test_datatypes_")
        fn = os.path.join(tmpdir, "datatypes.py")
        f = open(fn, "w")
        f.write(TEST_DATATYPE_SOURCE)
        f.close()
        registry = ZConfig.datatypes.Registry()

        # we really want the temp area to override everything else:
        sys.path.insert(0, tmpdir)
        try:
            datatype = registry.get("datatypes.my_sample_datatype")
        finally:
            shutil.rmtree(tmpdir)
            sys.path[:] = old_sys_path
        self.assertEqual(datatype, 42)

    def test_register_shadow(self):
        reg = ZConfig.datatypes.Registry()
        self.assertRaisesRegex(ValueError,
                               "conflicts with built-in type",
                               reg.register,
                               'integer', None)

        reg.register("foobar", None)
        self.assertRaisesRegex(ValueError,
                               "already registered",
                               reg.register,
                               'foobar', None)

    def test_get_fallback_basic_key(self):
        reg = ZConfig.datatypes.Registry({})
        self.assertIsNone(reg._basic_key)
        self.assertRaisesRegex(ValueError,
                               "unloadable datatype name",
                               reg.get,
                               'integer')
        self.assertIsNotNone(reg._basic_key)


TEST_DATATYPE_SOURCE = """
# sample datatypes file

my_sample_datatype = 42
"""



##############################################################################
#
# Copyright (c) 2007 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""\
Support for working with ZConfig data without a schema.

"""
__docformat__ = "reStructuredText"

import ZConfig.cfgparser


def loadConfigFile(file, url=None):
    c = Context()
    Parser(Resource(file, url), c).parse(c.top)
    return c.top


class Resource:

    def __init__(self, file, url=''):
        self.file, self.url = file, url


class Section(dict):

    imports = ()

    def __init__(self, type='', name='', data=None, sections=None):
        dict.__init__(self)
        if data:
            self.update(data)
        self.sections = sections or []
        self.type, self.name = type, name

    def addValue(self, key, value, *args):
        if key in self:
            self[key].append(value)
        else:
            self[key] = [value]

    def __str__(self, pre=''):
        result = []

        if self.imports:
            for pkgname in self.imports:
                result.append('%import ' + pkgname)
            result.append('')

        if self.type:
            if self.name:
                start = f'{pre}<{self.type} {self.name}>'
            else:
                start = f'{pre}<{self.type}>'
            result.append(start)
            pre += '  '

        lst = sorted(self.items())
        for name, values in lst:
            for value in values:
                result.append(f'{pre}{name} {value}')

        if self.sections and self:
            result.append('')

        for section in self.sections:
            result.append(section.__str__(pre))

        if self.type:
            pre = pre[:-2]
            result.append(f'{pre}</{self.type}>')
            result.append('')

        result = '\n'.join(result).rstrip()
        if not pre:
            result += '\n'
        return result


class Context:

    def __init__(self):
        self.top = Section()
        self.sections = []

    def startSection(self, container, type_, name):
        newsec = Section(type_, name)
        container.sections.append(newsec)
        return newsec

    def endSection(self, container, type_, name, newsect):
        pass

    def importSchemaComponent(self, pkgname):
        if pkgname not in self.top.imports:
            self.top.imports += (pkgname, )

    def includeConfiguration(self, section, newurl, defines):
        raise NotImplementedError('includes are not supported')


class Parser(ZConfig.cfgparser.ZConfigParser):

    def handle_define(self, section, rest):
        raise NotImplementedError('defines are not supported')



##############################################################################
#
# Copyright (c) 2002, 2003, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""urlparse-like helpers that normalize file: URLs.

ZConfig and urllib2 expect file: URLs to consistently use the '//'
hostpart separator; the functions here enforce this constraint.

"""

import urllib.parse
import urllib.request


def urlnormalize(url):
    lc = url.lower()
    if lc.startswith("file:/") and not lc.startswith("file:///"):
        url = "file://" + url[5:]
    return url


def urlunsplit(parts):
    parts = list(parts)
    parts.insert(3, '')
    url = urllib.request.urlunparse(tuple(parts))
    if (parts[0] == "file"
            and url.startswith("file:/")
            and not url.startswith("file:///")):
        # It may not be possible to get here anymore with
        # modern urlparse, at least not on posix?
        url = "file://" + url[5:]  # pragma: no cover
    return url


def urldefrag(url):
    url, fragment = urllib.parse.urldefrag(url)
    return urlnormalize(url), fragment


def urljoin(base, relurl):
    url = urllib.request.urljoin(base, relurl)
    if url.startswith("file:/") and not url.startswith("file:///"):
        # It may not be possible to get here anymore with
        # modern urlparse, at least not on posix?
        url = "file://" + url[5:]  # pragma: no cover
    return url



##############################################################################
#
# Copyright (c) 2019 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""\
Tests for ZConfig.pygments.ZConfigLexer.

"""


import unittest


try:
    import pygments.lexer
    import pygments.token
except ModuleNotFoundError:
    pygments = None
else:
    import ZConfig.pygments


class ZConfigLexerTestCase(unittest.TestCase):

    def setUp(self):
        if pygments is None:
            self.skipTest('pygments is not available')

    def test_comment(self):
        expected = [
            (pygments.token.Text, '  '),
            (pygments.token.Comment, '# some crazy text'),
            (pygments.token.Text, '\n'),
        ]
        t = list(ZConfig.pygments.ZConfigLexer().get_tokens(
            '  # some crazy text\n'))
        self.assertEqual(t, expected)

    def test_option_without_substitutions(self):
        expected = [
            (pygments.token.Name, 'logdir'),
            (pygments.token.Whitespace, ' \t'),
            (pygments.token.String, '/var/log/myapp'),
            (pygments.token.Text, '\n'),
        ]
        t = list(ZConfig.pygments.ZConfigLexer().get_tokens(
            'logdir \t/var/log/myapp\n'))
        self.assertEqual(t, expected)

    def test_option_with_substitutions(self):
        expected = [
            (pygments.token.Name, 'logdir'),
            (pygments.token.Whitespace, ' \t'),
            (pygments.token.String, '/var/log/'),
            (pygments.token.String.Escape, '$$'),
            (pygments.token.String, '-'),
            (pygments.token.String.Interpol, '$'),
            (pygments.token.Name, 'conf'),
            (pygments.token.String, '-'),
            (pygments.token.String.Interpol, '${'),
            (pygments.token.Name, 'fnoc'),
            (pygments.token.String.Interpol, '}'),
            (pygments.token.String, '-monthly'),
            (pygments.token.Text, '\n'),
        ]
        t = list(ZConfig.pygments.ZConfigLexer().get_tokens(
            'logdir \t/var/log/$$-$conf-${fnoc}-monthly\n'))
        self.assertEqual(t, expected)

    def test_named_section(self):
        expected = [
            (pygments.token.Name.Tag, '<sect'),
            (pygments.token.Whitespace, ' '),
            (pygments.token.Name, 'name'),
            (pygments.token.Name.Tag, '>'),
            (pygments.token.Text, '\n  '),
            (pygments.token.Name, 'option'),
            (pygments.token.Whitespace, ' '),
            (pygments.token.String, 'value'),
            (pygments.token.Text, '\n'),
            (pygments.token.Name.Tag, '</sect'),
            (pygments.token.Name.Tag, '>'),
            (pygments.token.Text, '\n'),
        ]
        text = ('<sect name>\n'
                '  option value\n'
                '</sect>\n')
        t = list(ZConfig.pygments.ZConfigLexer().get_tokens(text))
        self.assertEqual(t, expected)

    def test_unnamed_section(self):
        expected = [
            (pygments.token.Name.Tag, '<sect'),
            (pygments.token.Name.Tag, '>'),
            (pygments.token.Text, '\n  '),
            (pygments.token.Name, 'option'),
            (pygments.token.Whitespace, ' '),
            (pygments.token.String, 'value'),
            (pygments.token.Text, '\n'),
            (pygments.token.Name.Tag, '</sect'),
            (pygments.token.Name.Tag, '>'),
            (pygments.token.Text, '\n'),
        ]
        text = ('<sect>\n'
                '  option value\n'
                '</sect>\n')
        t = list(ZConfig.pygments.ZConfigLexer().get_tokens(text))
        self.assertEqual(t, expected)

    def test_empty_named_section(self):
        expected = [
            (pygments.token.Name.Tag, '<sect'),
            (pygments.token.Whitespace, ' '),
            (pygments.token.Name, 'name'),
            (pygments.token.Name.Tag, '/>'),
            (pygments.token.Text, '\n'),
        ]
        t = list(ZConfig.pygments.ZConfigLexer().get_tokens('<sect name/>'))
        self.assertEqual(t, expected)

    def test_empty_unnamed_section(self):
        expected = [
            (pygments.token.Name.Tag, '<sect/'),
            (pygments.token.Name.Tag, '>'),
            (pygments.token.Text, '\n'),
        ]
        t = list(ZConfig.pygments.ZConfigLexer().get_tokens('<sect/>'))
        self.assertEqual(t, expected)

    def test_define_without_substitutions(self):
        expected = [
            (pygments.token.Keyword, '%define'),
            (pygments.token.Whitespace, ' '),
            (pygments.token.Name, 'logdir'),
            (pygments.token.Whitespace, ' \t'),
            (pygments.token.String, '/var/log/myapp'),
            (pygments.token.Text, '\n'),
        ]
        t = list(ZConfig.pygments.ZConfigLexer().get_tokens(
            '%define logdir \t/var/log/myapp\n'))
        self.assertEqual(t, expected)

    def test_define_with_substitutions(self):
        expected = [
            (pygments.token.Keyword, '%define'),
            (pygments.token.Whitespace, ' '),
            (pygments.token.Name, 'logdir'),
            (pygments.token.Whitespace, ' \t'),
            (pygments.token.String, '/var/log/'),
            (pygments.token.String.Escape, '$$'),
            (pygments.token.String, '-'),
            (pygments.token.String.Interpol, '$'),
            (pygments.token.Name, 'conf'),
            (pygments.token.String, '-'),
            (pygments.token.String.Interpol, '${'),
            (pygments.token.Name, 'fnoc'),
            (pygments.token.String.Interpol, '}'),
            (pygments.token.String, '-monthly'),
            (pygments.token.Text, '\n'),
        ]
        t = list(ZConfig.pygments.ZConfigLexer().get_tokens(
            '%define logdir \t/var/log/$$-$conf-${fnoc}-monthly\n'))
        self.assertEqual(t, expected)

    def test_import_without_substitutions(self):
        expected = [
            (pygments.token.Keyword, '%import'),
            (pygments.token.Whitespace, ' '),
            (pygments.token.String, 'ZConfig.components.logger'),
            (pygments.token.Text, '\n'),
        ]
        t = list(ZConfig.pygments.ZConfigLexer().get_tokens(
            '%import ZConfig.components.logger\n'))
        self.assertEqual(t, expected)

    def test_import_with_substitutions(self):
        expected = [
            (pygments.token.Keyword, '%import'),
            (pygments.token.Whitespace, ' '),
            (pygments.token.String, 'ZConfig.components.'),
            (pygments.token.String.Interpol, '${'),
            (pygments.token.Name, 'module'),
            (pygments.token.String.Interpol, '}'),
            (pygments.token.Text, '\n'),
        ]
        t = list(ZConfig.pygments.ZConfigLexer().get_tokens(
            '%import ZConfig.components.${module}\n'))
        self.assertEqual(t, expected)

    def test_include_without_substitutions(self):
        expected = [
            (pygments.token.Keyword, '%include'),
            (pygments.token.Whitespace, ' '),
            (pygments.token.String, 'somefile.conf'),
            (pygments.token.Text, '\n'),
        ]
        t = list(ZConfig.pygments.ZConfigLexer().get_tokens(
            '%include somefile.conf\n'))
        self.assertEqual(t, expected)

    def test_include_with_substitutions(self):
        expected = [
            (pygments.token.Keyword, '%include'),
            (pygments.token.Whitespace, ' '),
            (pygments.token.String.Interpol, '$'),
            (pygments.token.Name, 'basename'),
            (pygments.token.String, '.conf'),
            (pygments.token.Text, '\n'),
        ]
        t = list(ZConfig.pygments.ZConfigLexer().get_tokens(
            '%include $basename.conf\n'))
        self.assertEqual(t, expected)



##############################################################################
#
# Copyright (c) 2017, 2018 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import argparse
import itertools
import sys
import textwrap
from abc import ABC
from abc import abstractmethod
from itertools import filterfalse

import ZConfig.loader
from ZConfig.datatypes import null_conversion
from ZConfig.info import AbstractType
from ZConfig.info import MultiKeyInfo
from ZConfig.info import SectionInfo
from ZConfig.info import SectionType
from ZConfig.info import ValueInfo


MARKER = object()


class _VisitorBuilder:

    def __init__(self):
        self.visitors = []

    def __call__(self, Type):
        def dec(func):
            self.visitors.append((Type, func))
            return func
        return dec


class AbstractSchemaFormatter(ABC):

    def __init__(self, schema, stream=None):
        self.stream = stream or sys.stdout
        self._dt = schema.registry.find_name

    def write(self, *args):
        print(*args, file=self.stream)

    @abstractmethod
    def esc(self, x):
        "Escape blocks of text if needed"

    def _dedent(self, text):
        # dedent the text to avoid producing unwanted
        # definition lists. The XML parser strips leading whitespace from
        # the first line, but preserves it for subsequent lines, so for dedent
        # to work we have to ignore that first line.
        texts = text.split("\n")
        if len(texts) > 1:
            trail = textwrap.dedent('\n'.join(texts[1:]))
            text = texts[0] + '\n' + trail
        return text

    @abstractmethod
    def item_list(self):
        "Context manager for listing description items"

    def _describing(self, description, after):
        if description is not MARKER:
            with self.described_as():
                self.description(description)
                if after:
                    after()

    @abstractmethod
    def describing(self, description=MARKER, after=None):
        "description term, optional body"

    def describing_name(self, concrete_name,
                        description=MARKER, datatype=None,
                        **kwargs):
        with self.describing(description):
            self.concrete_name(concrete_name)
            self.datatype(datatype)

            for k, v in sorted(kwargs.items()):
                if v:
                    self.write(self.esc(f"({k}: {v})"))

    def description(self, description):
        if description:
            self.write(self.esc(description))

    example = description

    @abstractmethod
    def described_as(self):
        "Description body context manager"

    @abstractmethod
    def abstract_name(self, name):
        "Abstract name"

    @abstractmethod
    def concrete_name(self, *name):
        "Concrete name"

    @abstractmethod
    def concrete_section_name(self, *name):
        "Name of a section a user can type in a config"

    def datatype(self, datatype):
        self.write("(%s)" % self._dt(datatype))

    @abstractmethod
    def body(self):
        "Context manager for the whole document"


class AbstractSchemaPrinter(ABC):

    def __init__(self, schema, stream=None,
                 allowed_names=(), excluded_names=()):
        self.schema = schema
        stream = stream or sys.stdout
        self._explained = set()
        self._seen_typenames = set()
        self.fmt = self._schema_formatter(schema, stream)

        def _make_predicate(names):
            names = {x.lower() for x in names}

            def predicate(name_info):
                name, _ = name_info
                return name and name.lower() in names

            return predicate

        def _make_filter(names, filt):
            iter_all = self._iter_schema_items
            pred = _make_predicate(names)

            def it():
                return filt(pred, iter_all())

            return it

        if allowed_names:
            self._iter_schema_items = _make_filter(allowed_names, filter)

        if excluded_names:
            excluded_names = {x.lower() for x in excluded_names}
            self._iter_schema_items = _make_filter(excluded_names, filterfalse)
            self._included = lambda st: st.name not in excluded_names

    @abstractmethod
    def _schema_formatter(self, schema, stream):
        "Return a formatter"

    def _included(self, st):
        return True

    def _explain(self, st):
        if st.name in self._explained:
            return

        self._explained.add(st.name)

        self.fmt.description(st.description)
        if not self._included(st):
            return

        self.fmt.example(getattr(st, 'example', None))

        for sub in st.getsubtypenames():
            with self.fmt.item_list():
                self.visit(None, st.getsubtype(sub))

    def _iter_schema_items(self):
        def everything():
            return itertools.chain(self.schema.itertypes(),
                                   self.schema)
        # The abstract types tend to be the most important. Since we
        # only document a concrete type the first time we find it, and
        # we can find extensions of abstract types beneath the abstract
        # type which is itself buried under a concrete section, all the
        # different permutations would be only documented once under
        # that section. By exposing these first, they get documented at
        # the top-level, and each concrete section that uses the
        # abstract type gets a reference to it.

        def abstract_sections(base):
            for name, info in base:
                if isinstance(info, SectionInfo):
                    if info.sectiontype.isabstract():
                        yield name, info

                    # XXX: This isn't catching everything. Witness the
                    # relstorage component.
                elif isinstance(info, SectionType):
                    yield from abstract_sections(info)
        return itertools.chain(abstract_sections(everything()), everything())

    def printSchema(self):
        # side-effect of building may be printing
        self.buildSchema()

    def buildSchema(self):
        seen = set()  # prevent duplicates at the top-level
        # as we find multiple abstract types
        with self.fmt.body():
            with self.fmt.item_list():
                for name, info in self._iter_schema_items():
                    if info in seen:
                        continue
                    seen.add(info)
                    self.visit(name, info)

    TypeVisitor = _VisitorBuilder()
    visitors = TypeVisitor.visitors

    def visit(self, name, info):
        for t, f in self.visitors:
            if isinstance(info, t):
                f(self, name, info)
                break
        else:
            self._visit_default(name, info)

    @TypeVisitor(SectionType)
    def _visit_SectionType(self, name, info):
        if info.name in self._seen_typenames:
            return
        self._seen_typenames.add(info.name)
        with self.fmt.describing():
            if info.datatype is not null_conversion:
                self.fmt.concrete_section_name(info.name)
            else:
                self.fmt.abstract_name(info.name)
            self.fmt.datatype(info.datatype)

        with self.fmt.described_as():
            self.fmt.description(info.description)
            self.fmt.example(info.example)

            with self.fmt.item_list():
                for sub in info:
                    self.visit(*sub)

    @TypeVisitor(SectionInfo)
    def _visit_SectionInfo(self, name, info):
        st = info.sectiontype
        if st.isabstract():
            with self.fmt.describing(info.description,
                                     lambda: self._explain(st)):
                self.fmt.abstract_name(st.name)
                self.fmt.concrete_name(info.name)

        else:
            with self.fmt.describing():
                self.fmt.concrete_section_name(info.attribute, info.name)
                self.fmt.datatype(info.datatype)

            with self.fmt.described_as():
                with self.fmt.item_list():
                    for sub in info.sectiontype:
                        self.visit(*sub)

        self.fmt.example(info.example)

    @TypeVisitor(AbstractType)
    def _visit_AbstractType(self, name, info):
        with self.fmt.describing(info.description,
                                 lambda: self._explain(info)):
            self.fmt.abstract_name(info.name)

    def _visit_default(self, name, info):
        # KeyInfo or MultiKeyInfo
        default = info.getdefault()
        if isinstance(default, ValueInfo):
            default = default.value

        name = info.name
        if isinstance(info, MultiKeyInfo):
            name = name + " (*)"
        self.fmt.describing_name(name, info.description, info.datatype,
                                 default=default, metadefault=info.metadefault)

    del TypeVisitor


def load_schema(schema, package=None):
    """
    Load the *schema* and return the schema object.

    By default, *schema* is interpreted as a path on disk to a schema
    file.

    If *package* is set to a non-empty string, then *package* must
    name a Python package, and the file in *schema* will be loaded
    from that package. The *schema* can either refer to a component
    definition (e.g., ``components.xml``) or to a schema.
    """

    if not package:
        # A schema file
        schema_reader = argparse.FileType('r')(schema)
        return ZConfig.loader.loadSchemaFile(schema_reader)

    try:
        # A component in a package
        schema_template = (
            "<schema><import package='%s' file='%s' /></schema>"
            % (package, schema))
        from io import StringIO
        return ZConfig.loader.loadSchemaFile(StringIO(schema_template))
    except ZConfig.UnknownDocumentTypeError:
        # Ok, not parseable as a component. Try a simple schema.
        return ZConfig.loader.loadSchema(f'package:{package}:{schema}')



##############################################################################
#
# Copyright (c) 2017, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import textwrap
import unittest
from io import StringIO

import docutils
import docutils.frontend
import docutils.parsers.rst
import docutils.parsers.rst.directives
import docutils.utils

from ZConfig import schema2html
from ZConfig.sphinx import RstSchemaFormatter
from ZConfig.sphinx import SchemaToRstDirective
from ZConfig.tests import support


docutils.parsers.rst.directives.register_directive("zconfig",
                                                   SchemaToRstDirective)


def run_transform(*args):
    if '--out' not in args and '-o' not in args:
        buf = StringIO()
        with support.stdout_replaced(buf):
            schema2html.main(args)
        return buf
    return schema2html.main(args)  # pragma: no cover


if schema2html.RstSchemaPrinter:
    def run_transform_rst(*args):
        args += ('--format', 'xml')
        # Capture stderr to suppress junk like
        # description:3: (ERROR/3) Unknown interpreted text role "class".
        # description:3: (ERROR/3) Unknown interpreted text role "func".
        with support.stderr_replaced():
            return run_transform(*args)
else:
    def run_transform_rst(*args):
        pass


class TestSchema2HTML(unittest.TestCase):

    def test_schema_only(self):
        res = run_transform(support.input_file('simple.xml'))
        self.assertIn('</html>', res.getvalue())
        run_transform_rst(support.input_file('simple.xml'))

    @support.with_stdin_from_input_file('simple.xml')
    def test_schema_only_redirect(self):
        res = run_transform("-")
        self.assertIn('</html>', res.getvalue())

    def test_cover_all_schemas(self):
        for name in ('base-datatype1.xml',
                     'base-datatype2.xml',
                     'base-keytype1.xml',
                     'base-keytype2.xml',
                     'base.xml',
                     'library.xml',
                     'simplesections.xml',):
            res = run_transform(support.input_file(name))
            self.assertIn('</html>', res.getvalue())
            run_transform_rst(support.input_file(name))

    def test_html_section_example(self):
        name = 'simplesections.xml'
        res = run_transform(support.input_file(name))
        out = res.getvalue()
        self.assertIn('Section Example', out)
        self.assertIn('Multisection Example', out)

    def test_rst_section_example(self):
        name = 'simplesections.xml'
        res = run_transform_rst(support.input_file(name))
        out = res.getvalue()
        self.assertIn('Section Example', out)
        self.assertIn('Multisection Example', out)

    def test_cover_logging_components(self):
        res = run_transform('--package', 'ZConfig.components.logger')
        self.assertIn('eventlog', res.getvalue())
        run_transform_rst('--package', 'ZConfig.components.logger')


class TestRst(unittest.TestCase):

    def _parse(self, text):
        document = docutils.utils.new_document(
            "Schema",
            settings=docutils.frontend.OptionParser(
                components=(docutils.parsers.rst.Parser,)
            ).get_default_values())

        parser = docutils.parsers.rst.Parser()
        text = textwrap.dedent(text)
        with support.stderr_replaced():
            # Capture stderr to suppress junk like
            # description:3: (ERROR/3) Unknown interpreted text role "class".
            # description:3: (ERROR/3) Unknown interpreted text role "func".
            parser.parse(text, document)
        return document

    def test_parse_package(self):
        text = """
        Document
        ========
        .. zconfig:: ZConfig.components.logger

        """
        document = self._parse(text)
        doc_text = document.astext()
        # Check that it produced output
        self.assertIn("SMTPHandler", doc_text)
        self.assertIn("Example:", doc_text)

    def test_parse_package_file(self):
        text = """
        Document
        ========
        .. zconfig:: ZConfig.components.logger
            :file: base-logger.xml

        """
        document = self._parse(text)
        doc_text = document.astext()
        # Check that it produced output, limited to
        # just that one file.
        self.assertNotIn("SMTPHandler", doc_text)
        self.assertIn("base-logger", doc_text)
        self.assertIn("Base definition", doc_text)
        self.assertIn("Example:", doc_text)

    def test_parse_package_limited_names(self):
        text = """
        Document
        ========
        .. zconfig:: ZConfig.components.logger
            :members: syslog logfile
        """
        document = self._parse(text)
        doc_text = document.astext()

        # Check that it produced output, limited to
        # just that one part of the tree
        self.assertNotIn("SMTPHandler", doc_text)
        self.assertIn("syslog", doc_text)
        self.assertIn("SyslogHandlerFactory", doc_text)
        self.assertIn("FileHandlerFactory", doc_text)

    def test_parse_package_excluded_names(self):
        text = """
        Document
        ========
        .. zconfig:: ZConfig.components.logger
            :members: ZConfig.logger.base-logger
            :excluded-members: ZConfig.logger.handler
        """
        document = self._parse(text)
        doc_text = document.astext()

        # Check that it produced output, limited to
        # just that one part of the tree
        # In this case, the root base-logger, but the handlers subtree
        # was excluded.
        self.assertIn("zconfig.logger.base-logger", doc_text)
        self.assertNotIn("SMTPHandler", doc_text)
        self.assertNotIn("syslog", doc_text)
        self.assertNotIn("SyslogHandlerFactory", doc_text)
        self.assertNotIn("FileHandlerFactory", doc_text)

    def test_parse_package_schema(self):
        text = """
        Document
        ========
        .. zconfig:: ZConfig.tests
            :file: sphinx_test_schema.xml
        """
        document = self._parse(text)
        doc_text = document.astext()

        # Check that it produced output
        self.assertIn("EventLogFactory", doc_text)
        self.assertIn("<eventlog>", doc_text)

    def test_description_dedent(self):
        text = """No leading whitespace on this line.
        But this line has whitespace.
        As does this one.
        """
        written = []

        class FUT(RstSchemaFormatter):

            def __init__(self):
                pass

            def _parsed(self, text, _):
                return text

            def write(self, *texts):
                written.extend(texts)

        fut = FUT()
        fut.description(text)

        dedented = ("No leading whitespace on this line.\n"
                    "But this line has whitespace.\n"
                    "As does this one.\n")

        self.assertEqual(written[0], dedented)


def test_suite():
    return unittest.defaultTestLoader.loadTestsFromName(__name__)


if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



##############################################################################
#
# Copyright (c) 2003, 2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""Support for command-line overrides for configuration settings.

This module exports an extended version of the :class:`~.ConfigLoader`
class from the :mod:`ZConfig.loader` module. This provides support for
overriding specific settings from the configuration file from the
command line, without requiring the application to provide specific
options for everything the configuration file can include.

Each setting is given by a value specifier string, as described by
:meth:`ExtendedConfigLoader.addOption`.

"""

import ZConfig
import ZConfig.loader
import ZConfig.matcher


class ExtendedConfigLoader(ZConfig.loader.ConfigLoader):
    """A :class:`~.ConfigLoader` subclass that adds support for
    command-line overrides.
    """

    def __init__(self, schema):
        ZConfig.loader.ConfigLoader.__init__(self, schema)
        self.clopts = []   # [(optpath, value, source-position), ...]

    def addOption(self, spec, pos=None):
        """Add a single value to the list of overridden values.

        The *spec* argument is a value specifier string of the form
        ``optionpath=value``. For example::

            some/path/to/key=value

        The *optionpath* specifies the "full path" to the
        configuration setting: it can contain a sequence of names,
        separated by ``/`` characters. Each name before the last names
        a section from the configuration file, and the last name
        corresponds to a key within the section identified by the
        leading section names. If *optionpath* contains only one name,
        it identifies a key in the top-level schema. *value* is a
        string that will be treated just like a value in the
        configuration file.

        A source position for the specifier may be given as *pos*. If
        *pos* is specified and not ``None``, it must be a sequence of
        three values. The first is the URL of the source (or some
        other identifying string). The second and third are the line
        number and column of the setting. These position information
        is only used to construct a :exc:`~.DataConversionError` when
        data conversion fails.
        """
        if pos is None:
            pos = "<command-line option>", -1, -1
        if "=" not in spec:
            e = ZConfig.ConfigurationSyntaxError(
                "invalid configuration specifier", *pos)
            e.specifier = spec
            raise e
        # For now, just add it to the list; not clear that checking
        # against the schema at this point buys anything.
        opt, val = spec.split("=", 1)
        optpath = opt.split("/")
        if "" in optpath:
            # // is not allowed in option path
            e = ZConfig.ConfigurationSyntaxError(
                "'//' is not allowed in an option path", *pos)
            e.specifier = spec
            raise e
        self.clopts.append((optpath, val, pos))

    def createSchemaMatcher(self):
        if self.clopts:
            sm = ExtendedSchemaMatcher(self.schema)
            sm.set_optionbag(self.cook())
        else:
            sm = ZConfig.loader.ConfigLoader.createSchemaMatcher(self)
        return sm

    def cook(self):
        return OptionBag(self.schema, self.schema, self.clopts)


class OptionBag:
    def __init__(self, schema, sectiontype, options):
        self.sectiontype = sectiontype
        self.schema = schema
        self.keypairs = {}
        self.sectitems = []
        self._basic_key = schema.registry.get("basic-key")
        for item in options:
            optpath, val, pos = item
            name = sectiontype.keytype(optpath[0])
            if len(optpath) == 1:
                self.add_value(name, val, pos)
            else:
                self.sectitems.append(item)

    def basic_key(self, s, pos):
        try:
            return self._basic_key(s)
        except ValueError as e:
            raise ZConfig.ConfigurationSyntaxError(
                f"could not convert basic-key value: {e}", *pos)

    def add_value(self, name, val, pos):
        if name in self.keypairs:
            L = self.keypairs[name]
        else:
            L = []
            self.keypairs[name] = L
        L.append((val, pos))

    def __contains__(self, name):
        return name in self.keypairs

    def get_key(self, name):
        """Return a list of (value, pos) items for the key 'name'.

        The returned list may be empty.
        """
        L = self.keypairs.get(name)
        if L:
            del self.keypairs[name]
            return L
        return []

    def keys(self):
        return self.keypairs.keys()

    def get_section_info(self, type_, name):
        L = []  # what pertains to the child section
        R = []  # what we keep
        for item in self.sectitems:
            optpath, val, pos = item
            s = optpath[0]
            bk = self.basic_key(s, pos)
            if name and self._normalize_case(s) == name:
                L.append((optpath[1:], val, pos))
            elif bk == type_:
                L.append((optpath[1:], val, pos))
            else:
                R.append(item)
        if L:
            self.sectitems[:] = R
            return OptionBag(self.schema, self.schema.gettype(type_), L)

    def finish(self):
        if self.sectitems or self.keypairs:
            raise ZConfig.ConfigurationError(
                "not all command line options were consumed")

    def _normalize_case(self, string):
        return string.lower()


class MatcherMixin:

    def set_optionbag(self, bag):
        self.optionbag = bag

    def addValue(self, key, value, position):
        try:
            realkey = self.type.keytype(key)
        except ValueError as e:
            raise ZConfig.DataConversionError(e, key, position)

        if realkey in self.optionbag:
            return
        ZConfig.matcher.BaseMatcher.addValue(self, key, value, position)

    def createChildMatcher(self, type_, name):
        sm = ZConfig.matcher.BaseMatcher.createChildMatcher(self, type_, name)
        bag = self.optionbag.get_section_info(type_.name, name)
        if bag is not None:
            sm = ExtendedSectionMatcher(
                sm.info, sm.type, sm.name, sm.handlers)
            sm.set_optionbag(bag)
        return sm

    def finish_optionbag(self):
        for key in list(self.optionbag.keys()):
            for val, pos in self.optionbag.get_key(key):
                ZConfig.matcher.BaseMatcher.addValue(self, key, val, pos)
        self.optionbag.finish()


class ExtendedSectionMatcher(MatcherMixin, ZConfig.matcher.SectionMatcher):
    def finish(self):
        self.finish_optionbag()
        return ZConfig.matcher.SectionMatcher.finish(self)


class ExtendedSchemaMatcher(MatcherMixin, ZConfig.matcher.SchemaMatcher):
    def finish(self):
        self.finish_optionbag()
        return ZConfig.matcher.SchemaMatcher.finish(self)



##############################################################################
#
# Copyright (c) 2002, 2003, 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Shell-style string substitution helper."""

import os
import re

import ZConfig


_name_re = r'[a-zA-Z_][a-zA-Z0-9_]*'
_name_match = re.compile(_name_re).match


def substitute(s, mapping):
    """Substitute values from *mapping* into *s*.

    *mapping* can be a :class:`dict` or any type that supports the
    ``get()`` method of the mapping protocol. Replacement values are
    copied into the result without further interpretation. Raises
    :exc:`~.SubstitutionSyntaxError` if there are malformed constructs
    in *s*.
    """

    if "$" in s:
        result = ''
        rest = s
        while rest:
            p, name, namecase, rest, vtype = _split(rest)
            result += p
            if name:
                v = None
                if vtype == 'define':
                    v = mapping.get(name)
                if vtype == 'env':
                    v = os.getenv(namecase)

                if v is None:
                    raise ZConfig.SubstitutionReplacementError(s, namecase)
                result += v
        return result
    else:
        return s


def isname(s):
    """Returns ``True`` if *s* is a valid name for a substitution
    text, otherwise returns ``False``.
    """

    m = _name_match(s)
    if m:
        return m.group() == s
    else:
        return False


def _split(s):
    # Return a four tuple:  prefix, name, namecase, suffix
    # - prefix is text that can be used literally in the result (may be '')
    # - name is a referenced name, or None
    # - namecase is the name with case preserved
    # - suffix is trailling text that may contain additional references
    #   (may be '' or None)
    if "$" in s:
        i = s.find("$")
        c = s[i + 1:i + 2]
        if c == "":
            raise ZConfig.SubstitutionSyntaxError(
                "illegal lone '$' at end of source")
        if c == "$":
            return s[:i + 1], None, None, s[i + 2:], None
        prefix = s[:i]
        vtype = 'define'
        if c == "{":
            m = _name_match(s, i + 2)
            if not m:
                raise ZConfig.SubstitutionSyntaxError(
                    "'${' not followed by name")
            name = m.group(0)
            i = m.end() + 1
            if not s.startswith("}", i - 1):
                raise ZConfig.SubstitutionSyntaxError(
                    "'${%s' not followed by '}'" % name)
        elif c == "(":
            m = _name_match(s, i + 2)
            if not m:
                raise ZConfig.SubstitutionSyntaxError(
                    "'$(' not followed by name")
            name = m.group(0)
            i = m.end() + 1
            if not s.startswith(")", i - 1):
                raise ZConfig.SubstitutionSyntaxError(
                    "'$(%s' not followed by ')'" % name)
            vtype = 'env'
        else:
            m = _name_match(s, i + 1)
            if not m:
                raise ZConfig.SubstitutionSyntaxError(
                    "'$' not followed by '$' or name")
            name = m.group(0)
            i = m.end()
        return prefix, name.lower(), name, s[i:], vtype
    else:
        return s, None, None, None, None



##############################################################################
#
# Copyright (c) 2003 Zope Corporation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""Script to check validity of a configuration file.
"""


import argparse
import sys

import ZConfig


def main(args=None):
    optparser = argparse.ArgumentParser(
        description="Script to check validity of a configuration file",
        epilog="""
        Each file named on the command line is checked for syntactical errors
        and schema conformance.  The schema must be specified.  If no files
        are specified and standard input is not a TTY, standard in is treated
        as a configuration file.  Specifying a schema and no configuration
        files causes the schema to be checked.""",
    )

    optparser.add_argument(
        "-s", "--schema", dest="schema",
        required=True,
        help="use the schema in FILE (can be a URL)",
        metavar="FILE"
    )

    optparser.add_argument(
        "file",
        nargs='*',
        help="Optional configuration file to check",
        type=argparse.FileType('r'),
    )

    options = optparser.parse_args(args=args)

    schema = ZConfig.loadSchema(options.schema)

    if not options.file:
        if sys.stdin.isatty():
            # just checking the schema
            return 0

        # stdin is a pipe
        options.file = [sys.stdin]

    errors = False
    for f in options.file:
        try:
            ZConfig.loadConfigFile(schema, f)
        except ZConfig.ConfigurationError as e:
            print(str(e), file=sys.stderr)
            errors = True

    return int(errors)


if __name__ == "__main__":
    sys.exit(main())
